# License for this source file
# 
# Copyright (c) 2014 Miqra Engineering
# 

import sys, os
# perform gstreamer imports (python3 style)
import gi
gi.require_version('Gst','1.0')

from gi.repository import GObject
from gi.repository import Gst


class BasicPlayer(GObject.GObject):
    __gsignals__ = {
    'playback-ready'          : (GObject.SIGNAL_RUN_LAST, GObject.TYPE_NONE,  # parameters: source_file tag_dict
                                (GObject.TYPE_STRING, GObject.TYPE_PYOBJECT)),   
    'playback-playing'        : (GObject.SIGNAL_RUN_LAST, GObject.TYPE_NONE,  # parameters: None
                                ()),   
    'playback-stopped'        : (GObject.SIGNAL_RUN_LAST, GObject.TYPE_NONE,  # parameters: None
                                ()),
    'playback-paused'         : (GObject.SIGNAL_RUN_LAST, GObject.TYPE_NONE,  # parameters: None
                                ()),
    'playback-finished'       : (GObject.SIGNAL_RUN_LAST, GObject.TYPE_NONE,  # parameters: None
                                ()),
    'playback-error'          : (GObject.SIGNAL_RUN_LAST, GObject.TYPE_NONE,  # parameters: player_state error debug
                                (GObject.TYPE_STRING, GObject.TYPE_STRING, GObject.TYPE_STRING,)),
    'volume-changed'          : (GObject.SIGNAL_RUN_LAST, GObject.TYPE_NONE,  # parameters: volume
                                (GObject.TYPE_FLOAT,)),     
    'playback-buffering'      : (GObject.SIGNAL_RUN_LAST, GObject.TYPE_NONE,  # parameters: volume
                                (GObject.TYPE_INT,)),     
                    }

    __gproperties__ = {

        'volume' : (GObject.TYPE_FLOAT,
                          'Volume',
                          'Volume',
                          0,
                          1,
                          1,
                          GObject.PARAM_READWRITE),
    }

    _running = False
    tags = {}

    source = None

    def __init__(self):
        GObject.GObject.__init__(self)
        # setup gstreamer
        self.pipeline_state = Gst.State.NULL;
        self.prepare_gstreamer()
        
        self.buffering = False

    def __del__(self):
        self.stop_gstreamer()

    def do_get_property(self, property):
        if property.name == 'volume':
            return self.pipeline.get_property('volume')
        else:
            raise AttributeError, 'unknown property %s' % property.name

    def do_set_property(self, property, value):
        if property.name == 'volume':
            volume = clamp(value,0.0,1.0)
            self.pipeline.set_property('volume', volume)
            #print "Set volume to {0}, got {1}".format(volume,self.pipeline.get_property('volume'))
            self.emit('volume-changed',volume)
        else:
            raise AttributeError, 'unknown property %s' % property.name

    def play(self):
        # start playback ;)
        if self.player_state in ["READY","PAUSED",]:
            self.player_state = "PLAYING"
            self.pipeline.set_state(Gst.State.PLAYING)
            self.emit('playback-playing')
        
    def load(self,file):
        self.source = "file://" + file
        #print "Attempting to load: '{0}'".format(file)
        self.pipeline.set_state(Gst.State.NULL);
        self.pipeline.set_property("uri", self.source)
        self.pipeline.set_state(Gst.State.PAUSED);
        
        self.player_state = "LOADING"
        self.tags.clear()

    def load_uri(self,uri):
        self.source = uri
        #print "Attempting to load: '{0}'".format(file)
        self.pipeline.set_state(Gst.State.NULL);
        self.pipeline.set_property("uri", self.source)
        self.pipeline.set_state(Gst.State.PAUSED);
        
        self.player_state = "LOADING"
        self.tags.clear()
    
    def stop(self):
        if self.player_state in ["PLAYING","PAUSED",]:
            self._running = False
            self.pipeline.set_state(Gst.State.READY)
            self.player_state = "READY"
            self.emit('playback-stopped')
            
    def pause(self,notify=True):
        # cannot pause/unpause if we're waiting for buffer
        if self.player_state in ["PLAYING",] and not self.buffering:
            self.player_state = "PAUSED"
            self.pipeline.set_state(Gst.State.PAUSED)
            if notify:
                self.emit('playback-paused')

    def _finished(self):
        self._running = False
        self.pipeline.set_state(Gst.State.READY)
        self.seek(0)
        self.player_state = "PAUSED"
        self.emit('playback-finished')


    def prepare_gstreamer(self):
        # bin containing the recorder stuff
        
        self.pipeline = Gst.ElementFactory.make("playbin", None)
        videosink = Gst.ElementFactory.make("glimagesink", None)
        alsasink = Gst.ElementFactory.make("alsasink", None)
        
        # set output device
        #devicename = self.config["Devices.Output"].getStr('name')
        #if common.check_alsadev(devicename):
        #    alsasink.set_property('device', devicename)
        
        self.pipeline.set_property("video-sink", videosink)
        self.pipeline.set_property("audio-sink", alsasink)
        
        # connect the bus listener to the message function
        self.bus = self.pipeline.get_bus()
        self.bus.add_signal_watch()
        self.bus.connect("message", self.on_message)

        self.player_state = "NONE"

    def stop_gstreamer(self):
        try:
            self.pipeline.get_bus().disconnect(self.busconnection)
            self.pipeline.get_bus().remove_signal_watch()
            self.pipeline.set_state(Gst.State.NULL)
        except GObject.GError, e:
            self.set_sensitive(True)    

    def seek(self,seconds):
        self.pipeline.seek_simple(Gst.Format.TIME, Gst.SeekFlags.FLUSH | Gst.SeekFlags.KEY_UNIT, seconds* Gst.SECOND)

    def duration(self):
        result = self.pipeline.query_duration(Gst.Format.TIME)
        if result is not None and result[0]:
            return float(result[1]) / Gst.SECOND
        else:
            return -1;
    
    def position(self):
        result = self.pipeline.query_position(Gst.Format.TIME)
        if result is not None and result[0]:
            return float(result[1]) / Gst.SECOND
        else:
            return -1;

    def on_message(self, bus, message):
        t = message.type
        
        # detect end of stream, and 
        if t == Gst.MessageType.EOS:
            self._finished()
        elif t == Gst.MessageType.ERROR:
            self.pipeline.set_state(Gst.State.NULL)
            err, debug = message.parse_error()
            #sys.stderr.write("Error: {0}\nDebug: {1}".format(err,debug))
            self.pipeline.set_state(Gst.State.NULL)
            self._running = False
            self.emit('playback-finished')
            self.emit('playback-error', self.player_state, err, debug)
            self.player_state = "NONE"
        elif t == Gst.MessageType.TAG:
            tags = message.parse_tag();
            for i in range(0,tags.n_tags()):
                key = tags.nth_tag_name(i)
                val = tags.get_value_index(key,0)
                self.tags[key] = val;
        elif t == Gst.MessageType.ASYNC_DONE:
            if message.src == self.pipeline:
                pass
        elif t == Gst.MessageType.STREAM_STATUS:
            (status,owner) = message.parse_stream_status()
#            print "Stream status: {0} (by {1})\n".format(status,owner)
            pass
        elif t == Gst.MessageType.BUFFERING:
            pct = message.parse_buffering()
            print "Buffering: {0}%".format(pct)
            self.emit('playback-buffering',pct)
            if pct != 100:
                if self.pipeline_state == Gst.State.PLAYING:
                    self.pipeline.set_state(Gst.State.PAUSED)
                self.buffering = True
            elif pct == 100:
                if self.player_state == "PLAYING":
                    self.pipeline.set_state(Gst.State.PLAYING)
                self.buffering = False
        elif t == Gst.MessageType.DURATION_CHANGED:
#            print "Stream duration changed: {0}\n".format(float(self.pipeline.query_duration(Gst.Format.TIME)[1])/Gst.SECOND)
            pass
        elif t == Gst.MessageType.STATE_CHANGED:
            if message.src == self.pipeline:
                (old,new,pending) = message.parse_state_changed()
                self.pipeline_state = new
#                print "State changed from '{0}' to '{1}' pending '{2}'\n".format(old,new,pending)
                if old == Gst.State.READY and new == Gst.State.PAUSED and self.player_state == "LOADING":            
                    self.pipeline.set_state(Gst.State.PAUSED)
                    self.player_state = "READY"
                    self.emit('playback-ready',self.source,dict(self.tags))
            

GObject.type_register(BasicPlayer)
